{  simg -- command line applications for deterministic image processing
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name avgset}
{@description Calculate average of an image set}
{@longdescription
Calculate average of an image set and corresponding pixel weight (=1/sqr(noise))
If DATAMIN and DATAMX attributes are found, they are used to mask out invalid pixels.}

{$I directives.pas}
uses simage,getparam;
var par               : tparams;
    ma                : TAAStr;
    cc                : CSImg_Config;
    fni,fno,fnm,fnn   : ansistring;
    iset              : CSImg_ImageSet;
    fi,fn             : TSImg_FPImage;
    mi,mi2            : CSImg_MetaInfo;
    d                 : boolean;
    err               : ansistring;
    i                 : longint;
    rn,gain,dmin,dmax : double;

begin
    with par do
    begin
        init;
        paramusage:='<infiles> [<outfile>]';
        fni:='';       defparam('<infiles>','Input images',@fni,tpastr,rapdreq);
                       defparam('-h','Help',helpf);
        fno:='';       defparam('<outfile>','Output average image (''-'' for stdout)',fno);
        fnn:='';       defparam('-n','Output square noise image',fnn);          deffmt('-n','<outfile>');
        fnm:='';       defparam('-pm','Output bad pixel mask -pm',fnm);         deffmt('-pm','<outfile>');
        d:=false;      defparam('-d','Drop brightest and darkest pixel',d);
        ma.size:=0;    par.defparam('-m','Add meta info (can be defined multiple times)',ma);   par.deffmt('-m','<key>=<value>[/<comment>]');
        run;
        done;
    end;
    cc:=CSImg_Config.create(SImg_Config_Default);
    cc.debug('avgset: src='+fni+' avg='+fno+' mask'+fnm+' noise='+fnn+' d='+boolean2str(d));

    fi.pt:=nil; fn.pt:=nil;
    iset:=CSImg_ImageSet_Files.init(cc,fni);
    mi:=CSImg_MetaInfo.create;
    iset.getimage(cc,fi,mi,0);

    { meta infos }
    for i:=0 to ma.size-1 do
    begin
        err:=mi.add(trim(ma.aa[i]));
        if err<>'' then  cc.warning(cc.name+': Error parsing -m '+ma.aa[i]+': '+err+ '  (expected syntax: -m <key>=<value>[/<comment>])');
    end;
    if not mi.get('DATAMIN',dmin) then dmin:=-1e300;
    if not mi.get('DATAMAX',dmax) then dmax:=1e300;

    { calculate and write avg, square noise }
    SImg_AvgSet(cc,iset, fi,fn, d, dmin,dmax);

    mi2:=CSImg_MetaInfo.create(mi);
    mi2.delete('DATAMIN');
    mi2.delete('DATAMAX');
    mi2.delete('RDNOISE');
    if fno<>'' then
    begin
        if fno='-' then fno:='';
        cc.Info('avgset: Writing average image '+fno);
        SImg_WriteImageFile(cc,fno,fi,mi2);
    end;
    mi2.delete('BUNIT');
    mi2.delete('EGAIN');
    if fnn<>'' then
    begin
        cc.Info('avgset: Writing noise image '+fnn);
        SImg_WriteImageFile(cc,fnn,fn,mi2);
    end;

    if fnm<>'' then
    begin
        if not mi.get('RDNOISE',rn) then cc.error('avgset: Read out noise not defined. Required for calculating pixel mask');
        if not mi.get('EGAIN',gain) then
        begin
            cc.warning('avgset: Gain not defined, assuming 1 (image unit = electrons)');
            gain:=1;
        end;
        SImg_GetPixelMask(cc, fi,fn,fn, rn,gain);
        cc.Info('avgset: Writing pixel mask image '+fnm);
        SImg_WriteImageFile(cc,fnm,fn,mi2);
    end;

    SImg_Free(cc,fi);
    SImg_Free(cc,fn);
    mi.destroy;
    mi2.destroy;;

    iset.done;
    cc.destroy;
end.
