{  simg -- command line applications for deterministic image processing
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name cropneg}
{@description Crop negative values and optimally regularize it using sqr(f(x))/(f(x)+r)}

{$I directives.pas}
uses simage,getparam,sysutils;
var par                 : tparams;
    cc                  : CSImg_Config;
    ifn,ofn             : ansistring;
    fi                  : TSImg_FPImage;
    a,b,q               : double;
    i                   : longint;
    mi                  : CSImg_MetaInfo;

begin
    cc:=CSImg_Config.create(SImg_Config_Default);

    with par do
    begin
        init;
        paramusage:='[<infile>] [<outfile>]';
        ifn:='';        defparam('<infile>', 'Source image (default: stdin)', ifn);
        ofn:='';        defparam('<outfile>','Destination image (default: stdout)', ofn);
        a:=0;           defparam('-a','Absolute regularization parameter: output is sqr(f(x))/(f(x)+r) where f(x) is the input and r is the sum of values given by -a and -q. Default: 0', a);
        q:=-1;          defparam('-q','Regularization parameter as quantile. output is sqr(f(x))/(f(x)+r) where f(x) is the input and r is the sum of values given by -a and -q. Default: 0', q);
                        defparam('-h','Help',helpf);
                        defparam('-o','Output image',ofn); deffmt('-o','<outfile>');
        run;
        done;
    end;
    cc.debug('cropneg: source='+ifn + ' dest='+ofn+' a='+real2str(a)+' q='+real2str(q));

    fi.pt:=nil;
    cc.Info('cropneg: Reading input image '+ifn);
    mi:=CSImg_MetaInfo.create;
    SImg_ReadImageFile(cc,ifn,fi,mi);

    cc.Info('cropneg: Crop negative values');
    if (q>=0) and (q<=1) then a+=SImg_GetQuantile(cc, fi, q);
    if a>simg_mina then
    begin
        cc.Info('cropneg: Crop negative values and regularize');
        for i:=0 to fi.size-1 do
        begin
            b:=fi.pt[i];
            if b<=0 then fi.pt[i]:=0
            else fi.pt[i]:=sqr(b)/(b+a);
        end;
    end
    else begin
        cc.Info('cropneg: Crop negative values');
        for i:=0 to fi.size-1 do
            if fi.pt[i]<0 then fi.pt[i]:=0
    end;

    cc.Info('cropneg: Writing output image '+ofn);
    SImg_WriteImageFile(cc,ofn,fi,mi);
    SImg_Free(cc,fi);

    cc.destroy;
end.
