{  simg -- command line applications for deterministic image processing
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name dcn}
{@description Deconvolution and denoising}

{$I directives.pas}
uses simage,getparam,sysutils;
var par                 : tparams;
    cc                  : CSImg_Config;
    ifn,ofn             : ansistring;
    fi                  : TSImg_FPImage;
    ifwhm,ofwhm,fwhmf   : double;
    dn,snr,bg,ddn,bdn   : double;
    ns,ss               : boolean;
    mi                  : CSImg_MetaInfo;

begin
    cc:=CSImg_Config.create(SImg_Config_Default);

    with par do
    begin
        init;
        paramusage:='[<infile>] [<outfile>]';
        ifn:='';        defparam('<infile>', 'Source image (default: stdin)', ifn);
        ofn:='';        defparam('<outfile>','Destination image (default: stdout)', ofn);
        ifwhm:=-1;      defparam('-fi','Input FWHM (default: auto-detect)',ifwhm);
        ofwhm:=1.5;     defparam('-fo','Output FWHM. Default: 1.5',ofwhm);
        fwhmf:=0.75;    defparam('-ff','Output FWHM factor. Default 0.75. Maximum of value defined by -fo and ff is used.',fwhmf);
        dn:=1;          defparam('-n','Denoising strength. 0..1. Default 1.',dn);
        ddn:=0;         defparam('-nd','Additional denoising of regions darker than that quantile. Default: 0',ddn);
        bdn:=0.01;      defparam('-nb','Additional denoising at boundaries. Relative to image dimensions. Default: 1%',bdn);
        bg:=1.0;        defparam('-b','Additional lower constraints. Recommended: 0..1.5. Default: 1. < 0.1 to Disable.',bg);
        ns:=false;      defparam('-ns','No star mode',ns);
        snr:=15;        defparam('-s','SNR parameter for wiener deconvolution. Default 15.',snr);
        ss:=false;      defparam('-ss','Suppress deconvolution of saturated areas. Should be enabled if image contains saturated stars.',ss);
                        defparam('-h','Help',helpf);
                        defparam('-o','Output image',ofn); deffmt('-o','<outfile>');
        run;
        done;
    end;
    cc.debug('dcn: source='+ifn + ' dest='+ofn+' ifwhm='+real2str(ifwhm)+' ofwhm='+real2str(ofwhm)+' fwhmf='+real2str(fwhmf)+' dn='+real2str(dn)+' ddn='+real2str(ddn)+' bdn='+real2str(bdn)+' bg='+real2str(bg)+' snr='+real2str(snr)
        +' ns='+boolean2str(ns)+' ss='+boolean2str(ss));

    fi.pt:=nil;
    cc.Info('dcn: Reading input image '+ifn);
    mi:=CSImg_MetaInfo.create;
    SImg_ReadImageFile(cc,ifn,fi,mi);

    if ifwhm<1e-10 then
    begin
        if mi.get('FWHMPIX',ifwhm) then cc.Info('dcn: Read PSF FWHM width from input image: ' +real2str(ifwhm,-1))
        else begin
            ifwhm:=SImg_GetSharpnessFWHM(cc,fi);
            cc.Info('dcn: Auto-detected input FWHM: ' +real2str(ifwhm,-1));
        end;
    end;
    fwhmf*=ifwhm;
    if fwhmf>ofwhm then ofwhm:=fwhmf;

    cc.Info('dcn: Deconvolution and denoising to output FWHM '+real2str(ofwhm,-1)+' using ' +int2str(cc.threads) + ' threads');
    SImg_DeconvDenoise(cc, fi,fi, ifwhm,ofwhm, dn,ddn,bdn, bg, ns, snr, ss);

    cc.Info('dcn: Writing output image '+ofn);
    mi.add('FWHMPIX',real2str(ofwhm,5),'Star FWHM in pixels');
    SImg_WriteImageFile(cc,ofn,fi,mi);
    SImg_Free(cc,fi);

    cc.destroy;
end.
