{  simg -- command line applications for deterministic image processing
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name drcompose}
{@description RGB composite from 1-color image with colorization according to dominating detail size.}

{$I directives.pas}
uses simage,getparam,sysutils;
var par                 : tparams;
    cc                  : CSImg_Config;
    img                 : TSImg_FPImage;
    mask                : TSImg_RDImage;
    ifn,ofn,mfn         : ansistring;
    l,sw,rs,rr,reg      : double;
    sa,ra               : array[0..2] of double;
    mi                  : CSImg_MetaInfo;

begin
    with par do
    begin
        init;
        paramusage:='[<infile> [<outfile>]]';
        ifn:='';        defparam('<infile>','Source image (default: stdin)',ifn);
        ofn:='';        defparam('<outfile>','Destination image (default: stdout)',ofn);
        mfn:='';        defparam('-mask','Mask image calculated by getmask', mfn);
        l:=0.05;        defparam('-l','Black level threshold as quantile or <0 to disbable it. Default: 5%',l);
        sw:=5;          defparam('-sw','Detail limit for white. Default 5. Should be at least 2 times the super-resolutuon factor.',sw);
        rs:=3;          defparam('-sr','Detail ratio between colors. Default 3.',rs);
        rr:=0.7;        defparam('-r','Dynamic range compression ratio between colors. Default: 0.7.',rr);
        reg:=2.0;       defparam('-n','Compression regularization relative to noise. Default: 2 (less compression/noise amplification for darker zones)', reg);
                        defparam('-o','Destination image',ofn); deffmt('-o','<outfile>');
                        defparam('-h','Help',helpf);
        run;
        done;
    end;

    cc:=CSImg_Config.create(SImg_Config_Default);
    cc.debug(cc.name+': source='+ifn+' dest='+ofn+' mask='+mfn+' l='+real2str(l)+' sw='+real2str(sw)+' rs='+real2str(rs)+' rr='+real2str(rr)+' reg='+real2str(reg));
    sa[0]:=sw;          ra[0]:=1;
    sa[1]:=sw*rs;       ra[1]:=rr;
    sa[2]:=sw*sqr(rs);  ra[2]:=sqr(rr);

    img.pt:=nil;
    mi:=CSImg_MetaInfo.create();
    cc.Info(cc.name+': Reading input image '+ifn);
    SImg_ReadImageFile(cc,ifn,img,mi);
    mi.deleteRaw;

    mask.pt:=nil;
    if mfn<>'' then
    begin
        cc.Info(cc.name+': Reading mask image '+mfn);
        SImg_ReadImageFile(cc,mfn,mask);
    end;

    cc.Info(cc.name+': High-pass filtering using '+int2str(cc.threads)+' threads');
    mi.add('NOISEFLR', real2str(SImg_DRCompose(cc, img,img,mask, l,reg, sa,ra),10),'Noise floor');

    cc.Info(cc.name+': Writing output image '+ofn);
    SImg_WriteImageFile(cc,ofn,img,mi);

    SImg_Free(cc,img);
    if mask.pt<>nil then SImg_Free(cc,mask);
    mi.destroy;
    cc.destroy;
end.
