{  simg -- command line applications for deterministic image processing
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name fourier}
{@description Fourier transforms an image}

{$I directives.pas}
uses simage,getparam;
var par        : tparams;
    cc         : CSImg_Config;
    img        : TSImg_FPImage;
    fi,fo      : ansistring;
    c,nopad    : boolean;

begin
    fi:='';
    fo:='';
    with par do
    begin
        init;
        paramusage:='[<infile> [<outfile>]]';
        fi:='';         defparam('<infile>','Source image (default: stdin)',fi);
        fo:='';         defparam('<outfile>','Destination image (default: stdout)',fo);
                        defparam('-h','Help',helpf);
        c:=false;       defparam('-c','Shift (0,0) to center (default: (0,0) is top left',c);
        nopad:=false;   defparam('-np','Disable padding',nopad);
                        defparam('-o','Destination image',fo); deffmt('-o','<outfile>');
        run;
        done;
    end;
    cc:=CSImg_Config.create(SImg_Config_Default);
    cc.debug('fourier: values: source='+fi+' dest='+fo+' center='+boolean2str(c)+' padding='+boolean2str(not nopad));

    cc.info('fourier: Reading input image '+fi);
    SImg_ReadImageFile(cc,fi,img);

    cc.info('fourier: FFT using '+int2str(cc.threads)+' threads');
    SImg_Spectrum(cc,img,img,c,not nopad);

    cc.info('fourier: Writing output image '+fo);

    SImg_WriteImageFile(cc,fo,img);
    SImg_Free(cc,img);
    cc.destroy;
end.
