{  simg -- command line applications for deterministic image processing
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name getgain}
{@description Calculate gain from image set}

{$I directives.pas}
uses simage,getparam;
var par         : tparams;
    ma          : TAAStr;
    cc          : CSImg_Config;
    fi,fbn      : ansistring;
    iset        : CSImg_ImageSet_Files;
    mi          : CSImg_MetaInfo;
    fb          : TSImg_FPImage;
    r           : double;
    i           : longint;


begin
    cc:=CSImg_Config.create(SImg_Config_Default);
    with par do
    begin
        init;
        paramusage:='<infiles>';
        fi:='';      defparam('<infiles>','Source images',@fi,nil,tpastr,rapdreq);
        fbn:='';     defparam('-b','Bias image',fbn);   deffmt('-b','<infile>');
        r:=-1;       defparam('-r','Read out noise',r);
                     defparam('-h','Help',helpf);
        ma.size:=0;  par.defparam('-m','Add meta info (can be defined multiple times)',ma);   par.deffmt('-m','<key>=<value>[/<comment>]');
        run;
        done;
    end;
    cc.debug('getnoise: src='+fi + ' rdnoise='+real2str(r)+' fbn='+fbn);
    mi:=CSImg_MetaInfo.create;

    { bias image }
    fb.pt:=nil;
    if fbn<>'' then
    begin
        cc.Info(cc.name+': Reading bias image '+fbn);
        SImg_ReadImageFile(cc,fbn,fb,mi);
    end;

    { meta infos }
    for i:=0 to ma.size-1 do
    begin
        fbn:=mi.add(trim(ma.aa[i]));
        if fbn<>'' then  cc.warning(cc.name+': Error parsing -m '+ma.aa[i]+': '+fbn+ '  (expected syntax: -m <key>=<value>[/<comment>])');
    end;
    if r>0 then mi.add('RDNOISE',real2str(r,5),'Read out noise in <image unit>''s');

    { get gain }
    iset:=CSImg_ImageSet_Files.init(cc,fi);
    r:=SImg_GetGain(cc,iset,fb,mi);
    writeln('Gain: ' + real2str(r,5) + ' e/ADU');

    SImg_Free(cc,fb);
    iset.done;
    cc.destroy;
end.
