{  simg -- image processing library
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{*************************************************************************}
{******** SImg_Write* ****************************************************}
{*************************************************************************}
{ Text alignment calculations
  a     : alignment parameter
  pos   : position input
  st    : size of the text
  si    : size of the image
  result: output position
}
function SImg_Write_align(a:single; pos,st,si:longint):longint;
var b   : boolean;
begin
    b:=a<0;
    a:=abs(a);
    if a>1 then a:=1;
    pos-=round(st*a);
    if b then
    begin
        if pos+st>si then pos:=si-st;
        if pos<0 then pos:=0;
    end;
    result:=pos;
end;

{******** SImg_Write12x22 ************************************************}
procedure SImg_Write12x22(cc:CSImg_Config; fi:TSImg_FPImage; x,y:longint; const s:shortstring);
var i,j,k      : longint;
    v          : word;
    d          : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi) then cc.error('SImg_Write12x22: Invalid image');
    x:=SImg_Write_align(cc.textHAlign, x, length(s)*12, fi.nx);
    y:=SImg_Write_align(cc.textVAlign, y, 22, fi.ny);

    for i:=1 to length(s) do
    begin
        for j:=0 to 21 do
        begin
            v:=SImg_Font12x22_data[ord(s[i]),j];
            if (y+j>=0) and (y+j<fi.ny) then
            begin
                d:=fi.pt + (y+j)*fi.sl + x*fi.nc;
                for k:=0 to 11 do
                begin
                    if (x+k>=0) and (x+k<fi.nx) then
                    begin
                        if v>=$8000 then move(cc.drawcolor, d^, fi.nc*sizeof(simgfp))
                        else if cc.TextBack then move(cc.backcolor, d^, fi.nc*sizeof(simgfp));
                        d+=fi.nc;
                    end;
                    v:=v shl 1;
                end;
            end;
        end;
        x+=12;
    end;
end;

{******** SImg_Write8x8 **************************************************}
procedure SImg_Write8x8(cc:CSImg_Config; fi:TSImg_FPImage;x,y:longint; const s:shortstring);
var i,j,k     : longint;
    v         : byte;
    d         : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi) then cc.Error('SImg_Write8x8: Invalid image');
    x:=SImg_Write_align(cc.textHAlign, x, length(s)*8, fi.nx);
    y:=SImg_Write_align(cc.textVAlign, y, 8, fi.ny);

    for i:=1 to length(s) do
    begin
        for j:=0 to 7 do
        begin
            v:=SImg_Font8x8_data[ord(s[i]),j];
            if (y+j>=0) and (y+j<fi.ny) then
            begin
                d:=fi.pt + (y+j)*fi.sl + x*fi.nc;
                for k:=0 to 7 do
                begin
                    if (x+k>=0) and (x+k<fi.nx) then
                    begin
                        if v>=$80 then move(cc.drawcolor, d^, fi.nc*sizeof(simgfp))
                        else if cc.TextBack then move(cc.backcolor, d^, fi.nc*sizeof(simgfp));
                        d+=fi.nc;
                    end;
                    v:=v shl 1;
                end;
            end;
        end;
        x+=8;
    end;
end;

{******** SImg_Write8x16 *************************************************}
procedure SImg_Write8x16(cc:CSImg_Config; fi:TSImg_FPImage;x,y:longint; const s:shortstring);
var i,j,k     : longint;
    v         : byte;
    d         : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi) then cc.Error('SImg_Write8x16: Invalid image');
    x:=SImg_Write_align(cc.textHAlign, x, length(s)*8, fi.nx);
    y:=SImg_Write_align(cc.textVAlign, y, 16, fi.ny);

    for i:=1 to length(s) do
    begin
        for j:=0 to 15 do
        begin
            v:=SImg_Font8x16_data[ord(s[i]),j];
            if (y+j>=0) and (y+j<fi.ny) then
            begin
                d:=fi.pt + (y+j)*fi.sl + x*fi.nc;
                for k:=0 to 7 do
                begin
                    if (x+k>=0) and (x+k<fi.nx) then
                    begin
                        if v>=$80 then move(cc.drawcolor, d^, fi.nc*sizeof(simgfp))
                        else if cc.TextBack then move(cc.backcolor, d^, fi.nc*sizeof(simgfp));
                        d+=fi.nc;
                    end;
                    v:=v shl 1;
                end;
            end;
        end;
        x+=8;
    end;
end;


{*************************************************************************}
{******** SImg_Line ******************************************************}
{*************************************************************************}
procedure SImg_Line(cc:CSImg_Config; fi:TSImg_FPImage; x1,y1,x2,y2:longint);
var i,j  : longint;
    d    : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi) then cc.error('SImg_Line: Invalid image');
    if abs(x2-x1)>=abs(y2-y1) then
    begin
        if x2<x1 then
        begin
            i:=x2; x2:=x1; x1:=i;
            i:=y2; y2:=y1; y1:=i;
        end;
        if x2=x1 then
        begin
            if (x1>=0) and (x1<fi.nx) and (y1>=0) and (y1<fi.ny) then
            begin
                d:=fi.pt + y1*fi.sl + x1*fi.nc;
                move(cc.drawcolor, d^, fi.nc*sizeof(simgfp));
            end;
         end
         else for i:=x1 to x2 do
         begin
            j:=round(((x2-i)*y1+(i-x1)*y2)/(x2-x1));
            if (i>=0) and (i<fi.nx) and (j>=0) and (j<fi.ny) then
            begin
                d:=fi.pt + j*fi.sl + i*fi.nc;
                move(cc.drawcolor, d^, fi.nc*sizeof(simgfp))
            end;
        end;
    end
    else begin
        if y2<y1 then
        begin
            i:=x2; x2:=x1; x1:=i;
            i:=y2; y2:=y1; y1:=i;
        end;
        for i:=y1 to y2 do
        begin
            j:=round(((y2-i)*x1+(i-y1)*x2)/(y2-y1));
            if (i>=0) and (i<=fi.ny) and (j>=0) and (j<=fi.nx) then
            begin
                d:=fi.pt + i*fi.sl + j*fi.nc;
                move(cc.drawcolor, d^, fi.nc*sizeof(simgfp));
            end;
        end;
    end;
end;


{*************************************************************************}
{******** SImg_Bar *******************************************************}
{*************************************************************************}
procedure SImg_Bar(cc:CSImg_Config; fi:TSImg_FPImage; x1,y1,x2,y2:longint);
var x,y,c  : longint;
    d      : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Integrity(fi) then cc.error('SImg_Bar: Invalid image');
    if x2<x1 then
    begin
        x:=x2;
        x2:=x1;
        x1:=x;
    end;
    if y2<y1 then
    begin
        y:=y2;
        y2:=y1;
        y1:=y;
    end;
    if x1<0 then x1:=0;
    if y1<0 then y1:=0;
    if x2>=fi.nx then x2:=fi.nx-1;
    if y2>=fi.ny then y2:=fi.ny-1;

    for y:=y1 to y2 do
    begin
        d:=fi.pt + y*fi.sl + x1*fi.nc;
        for x:=x1 to x2 do
            for c:=0 to fi.nc-1 do
            begin
                d^:=cc.drawcolor[c];
                d+=1;
             end;
    end;
end;

{*************************************************************************}
{******** SImg_Rectangle *************************************************}
{*************************************************************************}
procedure SImg_Rectangle(cc:CSImg_Config; fi:TSImg_FPImage; x1,y1,x2,y2:longint);
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi) then cc.error('SImg_Rectangle: Invalid Image');
    SImg_line(cc, fi, x1,y1,x2,y1);
    SImg_line(cc, fi, x2,y1,x2,y2);
    SImg_line(cc, fi, x2,y2,x1,y2);
    SImg_Line(cc, fi, x1,y2,x1,y1);
end;
