{  simg -- image processing library
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{*************************************************************************}
{*************************************************************************}
{******* SBIG3 (image format) routines ***********************************}
{*************************************************************************}
{*************************************************************************}
function SImg_RSBIG3(cc:CSImg_Config; fb:TSImg_FileBuf; var ri:TSImg_RDImage; mi:CSImg_MetaInfo):boolean;
var i,j,k            : longint;
    b,be,bn          : ^byte;
    ch               : ^char absolute b;
    w                : ^word absolute b;
    si               : ^shortint absolute b;
    e                : ^word;
    c,satd           : boolean;
    s,v              : shortstring;
    co,t,sat         : double;

begin
    if cc=nil then cc:=SImg_Config_Default;
    cc.debug('SImg_RSBIG3: called with: ri='+SImg_InfoStr2(ri)+'   '+int2str(fb.size)+' Bytes input buffer');
    SImg_RSBIG3:=false;

    if mi<>nil then mi.add('BUNIT','adu','Image unit');

    if fb.size>2048 then
    begin
        ri.nx:=0;
        ri.ny:=0;
        co:=-100;
        satd:=false;
        sat:=65535;  { makes the compiler happy }

        b:=fb.pt; be:=b+2048;
        s:='';
        while (b^>=32) and (b<be) do
        begin
            s+=ch^; b+=1;
        end;
        upstr(s);
        i:=length(s);
        if (i>=17) and (copy(s,i-16,17)=' COMPRESSED IMAGE') then c:=true
        else if (i>=6) and (copy(s,i-5,6)=' IMAGE') then c:=false
        else exit;

        repeat
            while (b^<=32) and (b<be) do
                b+=1;
            s:=''; v:='';
            while (b^<>61) and (b^>=32) and (b<be) do
            begin
                s+=ch^; b+=1;
            end;
            if b^=61 then
            begin
                b+=1;
                while (b^>=32) and (b<be) do
                begin
                    v+=ch^; b+=1;
                end;
                trimstr(v);
            end;
            trimstr(s);
            upstr(s);
            if s='HEIGHT' then
            begin
                val(v,ri.ny,i);
                if i<>0 then cc.warning('SImg_RSBIG3: Can''t evaluate value for HEIGHT');
            end;
            if s='WIDTH' then
            begin
                val(v,ri.nx,i);
                if i<>0 then cc.warning('SImg_RSBIG3: Can''t evaluate value for WIDTH');
            end;
            if s='PEDESTAL' then
            begin
                val(v,t,i);
                 if i<>0 then cc.warning('SImg_RSBIG3: Can''t evaluate value for PEDESTAL')
                  else co+=t;
            end;
            if (s='FILE_VERSION') and (v<>'3') then cc.warning('SImg_RSBIG3: File_version Should be 3');
            if (s='DATA_VERSION') and (v<>'1') then cc.warning('SImg_RSBIG3: Data_version Should be 1');
            if (s='EXPOSURE') and (mi<>nil) then mi.add('EXPTIME',v,'');
            if (s='SAT_LEVEL') then
            begin
                val(v,t,i);
                if i<>0 then cc.warning('SImg_RSBIG3: Can''t evaluate value for SAT_LEVEL')
                else begin
                    satd:=true;
                    sat:=t;
                end;
            end;
            if (s='E_GAIN') and (mi<>nil) then mi.add('EGAIN',v,'ADC resolution in electrons / ADU');
        until (s='END') or (b>=be);
        if b>=be then
        begin
            cc.warning('SImg_RSBIG3: Unexpected end of header section');
            exit;
        end;
        if ri.nx<1 then
        begin
            cc.warning('SImg_RSBIG3: WIDTH not defined');
            exit;
        end;
        if ri.ny<1 then
        begin
            cc.warning('SImg_RSBIG3: HEIGHT not defined');
            exit;
        end;

        SImg_New(cc,ri,SImg_RawWord,ri.nx,ri.ny,1,false);
        if satd and (mi<>nil) then mi.add('DATAMAX',real2str(sat+co,15),'Maximum valid value');
        ri.co:=co;

        b:=fb.pt+2048; be:=fb.pt+fb.size; bn:=b;
        e:=ri.ptw;
        k:=w^; { makes the compiler happy }
        if c then
            for i:=0 to ri.ny-1 do
            begin
                if b<be-1 then k:=w^; b+=2;
                if k>=ri.nx*2 then
                begin
                    for j:=0 to ri.nx-1 do
                    begin
                        if b<be-1 then e^:=w^;
                        b+=2; e+=1;
                     end;
                end
                else begin
                    bn:=b+k;
                    if b<bn-1 then k:=w^; b+=2;
                    e^:=k; e+=1;
                    for j:=1 to ri.nx-1 do
                        if b<bn then
                        begin
                            if si^=-128 then
                            begin
                                b+=1;
                                if b<bn-1 then k:=w^; b+=2;
                            end
                            else begin
                                k+=si^; b+=1;
                            end;
                            e^:=k; e+=1;
                        end;
                    b:=bn;
                end;
            end
        else begin
            be-=1;
            for i:=0 to ri.ny*ri.nx-1 do
            begin
                if b<be then e^:=w^;
                b+=2; e+=1;
            end;
            be+=1;
        end;
        if b>be then
        begin
            cc.warning('SImg_RSBIG3: Unexpected end of file');
            exit;
        end;
        SImg_RSBIG3:=true;
        if not SImg_Integrity(ri) then
        begin
            cc.Warning('SImg_RSBIG3: Internal error');
            SImg_RSBIG3:=false;
        end;
    end;
end;
