{  simg -- visualization library
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

const siv_image_count : longint = 0;

{ ************************************************************************* }
{ ******** CSIV_Image ***************************************************** }
{ ************************************************************************* }
{ ******** init *********************************************************** }
constructor CSIV_Image.init(fi:TSImg_FPImage;fn:simgstring);
begin
siv_debug_begin('CSIV_Image.init: called with fi='+SImg_InfoStr2(fi)+' fn='+fn);
    getmem2(histlimg,siv_histl_xsize*siv_hist_ysize*3+4);
    getmem2(histrimg,siv_histr_xsize*siv_hist_ysize*3+4);

    title:='';
    name:='';
    img16.pt:=nil;
    hist16c:=nil;
    histlc:=nil;
    histrc:=nil;

    marker_x:=-1;
    marker_y:=-1;
    siv_debug_end;

    load(fi,fn);
end;

{ ******** init *********************************************************** }
constructor CSIV_Image.init(anc:CSIV_Image);
var fi : TSImg_FPImage;
begin
    if anc=nil then siv_debug_begin('CSIV_Image.init')
    else siv_debug_begin('CSIV_Image.init: called with anc='+anc.name);
    getmem2(histlimg,siv_histl_xsize*siv_hist_ysize*3+4);
    getmem2(histrimg,siv_histr_xsize*siv_hist_ysize*3+4);
    img16.pt:=nil;
    hist16c:=nil;
    histlc:=nil;
    histrc:=nil;

    if anc=nil then
    begin
        marker_x:=-1;
        marker_y:=-1;
        siv_debug_end;
        fi.pt:=nil;
        load(fi,'');
    end
    else begin
        marker_x:=anc.marker_x;
        marker_y:=anc.marker_y;
        siv_debug_end;
        load(anc);
    end;
end;

{ ******** load *********************************************************** }
type CSIV_Image_load_thread = class(TThread)
            doStat,doDisc          : boolean;
            doneStat,doneDisc      : boolean;
            min,max,sum,ssum,delta : double;
            hist16c                : ^longint;
            constructor create(fi_p:TSImg_FPImage; ya_p,yz_p:longint);
            destructor destroy; override;
            procedure execute; override;
            procedure waitStat;
            procedure waitDisc;
            procedure startdisc(img16_p:TSimg_RDImage; min_p,max_p,delta_p:double);
        private
            fi        : TSImg_FPImage;
            img16     : TSimg_RDImage;
            ya,yn     : longint;
            procedure stat;
            procedure disc;
        end;

{ *** create *** }
constructor CSIV_Image_load_thread.create(fi_p:TSImg_FPImage; ya_p,yz_p:longint);
begin
    fi:=fi_p;
    ya:=ya_p;
    yn:=yz_p+1-ya_p;
    doStat:=true;
    doDisc:=false;
    doneStat:=false;
    doneDisc:=false;
    getmem2(hist16c,sizeof(longint)*65536*fi.nc);
    inherited create(false);
end;

{ *** destroy *** }
destructor CSIV_Image_load_thread.destroy;
begin
    while (not doneStat) or (not doneDisc) do
        delay(1);
    freemem2(hist16c,sizeof(longint)*65536*fi.nc);
    inherited destroy;
end;

{ *** execute *** }
procedure CSIV_Image_load_thread.execute;
begin
    repeat
        if doStat and (not doneStat) then
        begin
            stat;
            doneStat:=true;
        end
        else if doDisc and (not doneDisc) then
        begin
            disc;
            doneDisc:=true;
        end
        else delay(1);
    until doneStat and doneDisc;
end;

{ *** waitStat *** }
procedure CSIV_Image_load_thread.waitStat;
begin
    while not doneStat do
        delay(1);
end;

{ *** waitDisc *** }
procedure CSIV_Image_load_thread.waitDisc;
begin
    while not doneDisc do
        delay(1);
end;

{ *** startDisc *** }
procedure CSIV_Image_load_thread.startDisc(img16_p:TSimg_RDImage; min_p,max_p,delta_p:double);
begin
    min:=min_p;
    max:=max_p;
    delta:=delta_p;
    img16:=img16_p;
    doDisc:=true;
end;

{ *** stat *** }
procedure CSIV_Image_load_thread.stat;
var i,iz         : longint;
    r,s,t,v,w    : double;
    d            : ^simgfp;
begin
    { minval/maval }
    d:=fi.pt + ya*fi.sl;
    v:=d^; w:=v;
    s:=0; t:=0;
    iz:=fi.sl*yn-1;
    for i:=0 to iz do
    begin
        r:=d^; d+=1;
        if r<v then v:=r;
        if r>w then w:=r;
        s+=r; t+=sqr(r);
    end;
    min:=v;
    max:=w;
    sum:=s;
    ssum:=t;
end;

{ *** disc *** }
procedure CSIV_Image_load_thread.disc;
var i,j,k,iz,nc  : longint;
    r,s,t,m,del  : double;
    d            : ^simgfp;
    e            : ^word;
    h16c         : ^longint;
begin
    fillchar(hist16c^,sizeof(longint)*65536*img16.nc,0);
    h16c:=hist16c;
    d:=fi.pt + ya*fi.sl;
    e:=img16.ptw + ya*fi.sl;
    m:=min-(max-min)*1e-10;
    r:=1/(max-m);
    s:=(1-delta)*65535;
    iz:=fi.nx*yn-1;
    nc:=fi.nc;
    del:=delta;
    for i:=0 to iz do
        for k:=0 to nc-1 do
        begin
            t:=(d^-m)*r;
            j:=round(t*s/(1-del*t));
            h16c[j*nc+k]+=1;
            e^:=j;
            d+=1; e+=1;
        end;
end;

{ *** main *** }
procedure CSIV_Image.load(fi:TSImg_FPImage;fn:simgstring);
const maxthreads = 8;
var i,j,k,l,m,tn : longint;
    r,s,t        : double;
    time         : double;
    thr          : array[0..maxthreads-1] of CSIV_Image_load_thread;

begin
    siv_debug_begin('CSIV_Image.load: called with fi='+SImg_InfoStr2(fi)+' fn='+fn);
    time:=stimer;
    try
        SImg_Free(siv_config,img16);
        if hist16c<>nil then
        begin
            freemem2(hist16c);
            hist16c:=nil;
        end;
        if histlc<>nil then
        begin
            freemem2(histlc);
            histlc:=nil;
        end;
        if histrc<>nil then
        begin
            freemem2(histrc);
            histrc:=nil;
        end;

        if fi.pt=nil then SImg_QuiteException;
        if not SImg_Recover(siv_config,fi) then siv_error('CSIV_Image.load: Invalid image');

        SImg_New(siv_config, img16, SImg_RawWord, fi.nx,fi.ny,fi.nc, false);
        getmem2(hist16c,sizeof(longint)*65536*img16.nc);

        { minval/maval }
        tn:=siv_config.threads;
        if tn>maxthreads then tn:=maxthreads;
        for i:=0 to tn-1 do
        begin
            j:=(fi.ny*i) div tn;
            thr[i]:=CSIV_Image_load_thread.create(fi, (fi.ny*i) div tn, ((fi.ny*(i+1)) div tn)-1 );
        end;

        minv:=fi.pt^; maxv:=minv;
        s:=0; t:=0;
        for i:=0 to tn-1 do
        begin
            thr[i].waitStat;
            if thr[i].min<minv then  minv:=thr[i].min;
            if thr[i].max>maxv then  maxv:=thr[i].max;
            s+=thr[i].sum;
            t+=thr[i].ssum;
        end;
        s/=fi.size; mean:=s;
        t:=t/fi.size-sqr(s); { rounding errors }
        if t<0 then rms:=0
        else rms:=sqrt(t);
        if mean-minv<1e-100 then siv_error('CSIV_Image.load: Constant image');

        s:=1e-4*(maxv-minv);
        r:=mean;
        if r<minv+s then r:=minv+s;
        if r>maxv-s then r:=maxv-s;
        delta:=2-(maxv-minv)/(r-minv);

        { discretization }
        for i:=0 to tn-1 do
            thr[i].startDisc(img16, minv,maxv,delta);
        for i:=0 to tn-1 do
            thr[i].waitDisc;

        { histogram }
        dist16[0]:=0;
        j:=0;
        for i:=0 to 65535 do
        begin
            dist16[i+1]:=dist16[i];
            for k:=0 to img16.nc-1 do
            begin
                m:=0;
                for l:=0 to tn-1 do
                    m+=thr[l].hist16c[j];
                hist16c[j]:=m;
                dist16[i+1]+=m;
                j+=1;
            end;
        end;

        for i:=0 to tn-1 do
            thr[i].destroy;

        { setup default values }
        red:=0;
        green:=1 mod fi.nc;
        blue:=2 mod fi.nc;
        if fi.ny+fi.ny div 3>fi.nx then i:=fi.ny+fi.ny div 3
        else i:=fi.nx;
        if i>=1016 then
        begin
            scale:=i div 1016;
            if scale>siv_maxscale then scale:=siv_maxscale;
        end
        else if i<=192 then
        begin
            scale:=-(192 div i);
            if scale<-siv_maxscale then scale:=-siv_maxscale;
        end
        else scale:=0;
        inv:=false;  log:=false;
        ll:=0; ul:=65535;
        bet:=0; gam:=0;
        siv_image_count+=1;
        name:=fn;
        if fn='' then title:='(stdin) - '+int2str(siv_image_count)
        else title:=fn+' - '+int2str(siv_image_count);
        siv_global_history.insert(fn);

    except
        SIV_OutputCatchedException;
        img16.nx:=0; img16.ny:=0; img16.nc:=0;
        minv:=1; maxv:=0;
        mean:=0; rms:=0; meandev:=0;

        red:=0;  green:=0;  blue:=0;
        scale:=0;
        inv:=false;  log:=false;
        ll:=0; ul:=65535;
        bet:=0; gam:=0;
        name:='';
        title:='(empty)';
    end;
    siv_debug_end('CSIV_Image.load: performance: '+real2str((stimer-time)*1000,-2)+'ms ');
    writeln(stderr,(stimer-time)*1000:0:2,'ms');

    update_limits:=true;
    update_bg:=true;
    update_colors:=true;
    update_scale:=true;
{     update;   !!! run it manually !!! }
end;


{ ******** load *********************************************************** }
procedure CSIV_Image.load(anc:CSIV_Image);
var time         : double;

begin
    siv_debug_begin('CSIV_Image.load: called with anc='+anc.title);
    time:=stimer;
    minv:=anc.minv; maxv:=anc.maxv;
    mean:=anc.mean; rms:=anc.rms; meandev:=anc.meandev;
    delta:=anc.delta;

    red:=anc.red; green:=anc.green; blue:=anc.blue;
    scale:=anc.scale;
    inv:=anc.inv; log:=anc.log;
    ll:=anc.ll; ul:=anc.ul;
    bet:=anc.bet; gam:=anc.gam;
    name:=anc.name;

    try
        SImg_Free(siv_config,img16);
        if hist16c<>nil then
        begin
            freemem2(hist16c);
            hist16c:=nil;
        end;
        if histlc<>nil then
        begin
            freemem2(histlc);
            histlc:=nil;
        end;
        if histrc<>nil then
        begin
            freemem2(histrc);
            histrc:=nil;
        end;

        if not SImg_Recover(siv_config,anc.img16) then SImg_QuiteException;

        move(anc.dist16,dist16,sizeof(dist16));

        getmem2(hist16c,sizeof(longint)*65535*anc.img16.nc);
        move(anc.hist16c^,hist16c^,sizeof(longint)*65535*anc.img16.nc);
        SImg_New(siv_config, img16, SImg_RawWord, anc.img16.nx,anc.img16.ny,anc.img16.nc, false);
        move(anc.img16.pt^,img16.pt^,img16.dsize);

        siv_image_count+=1;
        if name='' then title:='(stdin) - '+int2str(siv_image_count)
        else title:=name+' - '+int2str(siv_image_count);

    except
        SIV_OutputCatchedException;
        img16.nx:=0; img16.ny:=0; img16.nc:=0;
        title:='(empty)';
    end;
    siv_debug_end('CSIV_Image.load: performance: '+real2str((stimer-time)*1000,-2)+'ms ');

    update_limits:=true;
    update_bg:=true;
    update_colors:=true;
    update_scale:=true;
    { update;   !!! run it manually !!! }
end;


{ ******** update ********************************************************* }
procedure CSIV_Image.update;
var i,j   : longint;
begin
    redraw_histl:=redraw_histl or update_limits or update_bg or update_colors;
    redraw_histr:=redraw_histr or update_limits or update_bg or update_colors;
    if update_limits or update_bg then img16_img8;
    if redraw_histl then drawhistl;
    if redraw_histr then drawhistr;
    if redraw_histl or redraw_histr then
    begin
        i:=marker_x; marker_x:=-1;
        j:=marker_y; marker_y:=-1;
        history_markers(i,j);
    end;
    update_scale:=false;
    update_colors:=false;
end;

{ ******** scaledWidth **************************************************** }
function CSIV_Image.scaledWidth:longint;
begin
    if scale>0 then result:=img16.nx div (scale+1)
    else result:=img16.nx*(1-scale);
end;

{ ******** scaledHeight *************************************************** }
function CSIV_Image.scaledHeight:longint;
begin
    if scale>0 then result:=img16.ny div (scale+1)
    else result:=img16.ny*(1-scale);
end;

{ ******** done *********************************************************** }
destructor CSIV_Image.done;
begin
    siv_debug_begin('CSIV_Image.done: Bye');
    freemem2(histlimg);
    freemem2(histrimg);
    SImg_Free(siv_config,img16);
    if hist16c<>nil then freemem2(hist16c);
    if histlc<>nil then freemem2(histlc);
    if histrc<>nil then freemem2(histrc);
    siv_debug_end;
end;


{ ******** private ******************************************************** }
{ ******** img16_img8 ****************************************************** }
procedure CSIV_Image.img16_img8;
var ill,ilu       : longint;
    so,sr,r65535  : double;
    f,g           : ^longint;
    i,j1,k        : longint;
    r,s           : double;
    alp1,alp2     : double;
    time          : double;
{    t1a,t1b,t2,t3 : double;}

begin
    siv_debug_begin('CSIV_Image.img16b_img8: bet='+real2str(bet)+' gam='+real2str(gam)+' inv='+boolean2str(inv));
    time:=stimer;
    r65535:=1/65535;
    try
        if not SImg_Recover(siv_config,img16) then SImg_QuiteException;
        if hist16c=nil then SImg_QuiteException;
        if histlc=nil then getmem2(histlc,sizeof(longint)*siv_histl_xsize*img16.nc);
        if histrc=nil then getmem2(histrc,sizeof(longint)*siv_histr_xsize*img16.nc);
        if ll<0 then ll:=0;
        if ul>65535 then ul:=65535;
        ill:=round(ll);
        ilu:=round(ul);
        if ill<ilu then
        begin
            ql:=dist16[ill]/dist16[65536];
            qu:=dist16[ilu+1]/dist16[65536];
        end
        else begin
            ql:=dist16[ill+1]/dist16[65536];
            qu:=dist16[ilu]/dist16[65536];
        end;
        if abs(ilu-ill)<2 then siv_error('Limits to close');
        so:=ill/(1-delta+delta*ill*r65535);
        sr:=1/(ilu/(1-delta+delta*ilu*r65535)-so);

        { alp1 }
        r:=gam-bet*0.5;
        if r<-0.9999 then r:=0.9999;
        if r> 0.9999 then r:=0.9999;
        r:=pi*(1-r)/4; alp1:=sin(r)/cos(r);

        { alp2 }
        r:=gam+bet*0.5;
        if r<-0.9999 then r:=0.9999;
        if r> 0.9999 then r:=0.9999;
        r:=pi*(1-r)/4; alp2:=cos(r)/sin(r);

        { xlat, histtrans, histrc, histlc }
        fillchar(histlc^,sizeof(longint)*siv_histl_xsize*img16.nc,0);
        fillchar(histrc^,sizeof(longint)*siv_histr_xsize*img16.nc,0);
        for i:=0 to 65535 do
        begin
            s:=(i/(1-delta+delta*i*r65535)-so)*sr;

            j1:=round(s*(siv_histr_xsize-1));
            k:=((siv_histl_xsize-siv_histr_xsize) shr 1) + j1;
            if (k>=0) and (k<siv_histl_xsize) then
            begin
                f:=histlc + k*img16.nc;
                g:=hist16c + i*img16.nc;
                for k:=0 to img16.nc-1 do
                begin
                    f^+=g^;
                    f+=1; g+=1;
                end;
            end;
            if j1<0 then
            begin
                if inv then k:=255
                else k:=0;
            end
            else if j1>=siv_histr_xsize then
            begin
                if inv then k:=0
                else k:=255;
            end
            else begin
                if ((i<=ill) and (ill<ilu)) or ((i>=ill) and (ill>ilu))  then s:=0
                else if ((i>=ilu) and (ilu>ill)) or ((i<=ilu) and (ilu<ill)) then s:=1
                else s:=exp(ln(1-exp(ln(1-s)*alp1))*alp2);
                if inv then s:=1-s;
                f:=histrc + round(s*(siv_histr_xsize-1))*img16.nc;
                g:=hist16c + i*img16.nc;
                for k:=0 to img16.nc-1 do
                begin
                    f^+=g^;
                    f+=1; g+=1;
                end;
                histtrans[j1]:=round(s*(siv_hist_ysize-1));
                k:=round(s*255);
            end;
            xlat[i]:=k;
        end;
    except
        SIV_OutputCatchedException;
    end;
    siv_debug_end('CSIV_Image.img16b_img8: performance: '+real2str((stimer-time)*1000,-2)+'ms ');
{    writeln(stderr,'t1a=',t1a-time:0:2,'  t1b=',t1b-t1a:0:2,'  t2=',t2-t1b:0:2,'  t3=',t3-t2:0:2,' t4=',(stimer-t3)*1000:0:2);}

    update_bg:=false;
    update_limits:=false;
end;

{ ******** getRGB ********************************************************* }
type CSIV_Image_toRGB_thread = class(CSImg_AutoThread)
            img            : CSIV_Image;
            buf            : pbyte;
            xa_,ya,xz_,yz  : longint;
            constructor create(p_img:CSIV_Image; p_buf:pbyte; p_xa,p_ya,p_xz,p_yz:longint);
            procedure execute; override;
        end;

constructor CSIV_Image_toRGB_thread.create(p_img:CSIV_Image; p_buf:pbyte; p_xa,p_ya,p_xz,p_yz:longint);
begin
    img:=p_img;
    buf:=p_buf;
    xa_:=p_xa;
    ya:=p_ya;
    xz_:=p_xz;
    yz:=p_yz;
    inherited create(true);
end;

procedure CSIV_Image_toRGB_thread.execute;
var i,j,k,x,y      : longint;
    xa,xz,sl,nc    : longint;
    red,green,blue : longint;
    lr,lg,lb       : longint;
    br,bg,bb       : byte;
    r,g,b,e,f,ptw  : ^word;
    d,xlat         : ^byte;

begin
    try
        xa:=xa_;
        xz:=xz_;
        nc:=img.img16.nc;
        sl:=img.img16.sl;
        ptw:=img.img16.ptw;
        red:=img.red mod nc;
        green:=img.green mod nc;
        blue:=img.blue mod nc;
        xlat:=@img.xlat;

        if img.scale=0 then
        begin
            d:=buf;
            for y:=ya to yz do
            begin
                e:=img.img16.ptw + y*sl + xa*nc;
                r:=e + red;
                g:=e + green;
                b:=e + blue;
                for x:=xa to xz do
                begin
                    d^:=xlat[r^]; d+=1; r+=nc;
                    d^:=xlat[g^]; d+=1; g+=nc;
                    d^:=xlat[b^]; d+=1; b+=nc;
                end;
            end;
        end
        else if img.scale>0 then
        begin
            i:=img.scale+1;
            d:=buf;
            for y:=ya to yz do
            begin
                e:=ptw + y*i*sl + xa*i*nc;
                for x:=xa to xz do
                begin
                    lr:=0; lg:=0; lb:=0;
                    f:=e; e+=i*nc;
                    for j:=1 to i do
                    begin
                        r:=f+red;
                        g:=f+green;
                        b:=f+blue;
                        for k:=1 to i do
                        begin
                            lr+=xlat[r^]; r+=nc;
                            lg+=xlat[g^]; g+=nc;
                            lb+=xlat[b^]; b+=nc;
                        end;
                        f+=sl;
                    end;
                    j:=i*i;
                    d[0]:=lr div j;
                    d[1]:=lg div j;
                    d[2]:=lb div j;
                    d+=3;
                end;
            end;
        end
        else begin { if img.scale<0 then }
            i:=-img.scale+1;
            d:=buf;
            for y:=ya to yz do
            begin
                e:=ptw + (y div i)*sl + (xa div i)*nc;
                r:=e + red;   br:=xlat[r^]; r+=nc;
                g:=e + green; bg:=xlat[g^]; g+=nc;
                b:=e + blue;  bb:=xlat[b^]; b+=nc;
                j:=i-(xa mod i);
                for x:=xa to xz do
                begin
                    if j=0 then
                    begin
                        br:=xlat[r^]; r+=nc;
                        bg:=xlat[g^]; g+=nc;
                        bb:=xlat[b^]; b+=nc;
                        j:=i;
                    end;
                    d[0]:=br;
                    d[1]:=bg;
                    d[2]:=bb;
                    d+=3;
                    j-=1;
                end;
            end;
        end;
    except
        SIV_OutputCatchedException;
    end;

    isrunning:=false;
    isrunning2:=false;
end;

procedure CSIV_Image.getRGB(var buf:pbyte; var x0,y0,width,height:longint);
const maxthreads = 8;
var i,j,tn  : longint;
    t       : array[0..maxthreads-1] of CSIV_Image_toRGB_thread;
begin
    try
        if not SImg_Recover(siv_config,img16) then SImg_QuiteException;

        if x0<0 then x0:=0;
        if y0<0 then y0:=0;
        i:=scaledWidth-x0;  if width>i then width:=i;
        i:=scaledHeight-y0; if height>i then height:=i;

        getmem2(buf,width*height*3);
//        writeln(stderr,'getRGB: ',x0,'  ',y0,'  ',width,'   ',height);

        tn:=siv_config.threads;
        if tn>maxthreads then tn:=maxthreads;
        for i:=0 to tn-1 do
        begin
            j:=(height*i) div tn;
            t[i]:=CSIV_Image_toRGB_thread.create(self, buf+j*width*3, x0, y0+j, x0+width-1, y0+((height*(i+1)) div tn)-1);
        end;
        for i:=0 to tn-1 do
            t[i].destroy;

    except
        SIV_OutputCatchedException;
    end;
end;

{ ******** drawhist16 ***************************************************** }
procedure CSIV_Image.drawhistl;
var i,j,k,l    : longint;
    k1,k2      : longint;
    r          : double;
    e          : ^byte;
    f          : ^longint;
//    time       : double;
begin
//    siv_debug_begin(-1,'CSIV_Image.drawhistl','');
//    time:=stimer;
    try
        if histlc=nil then SImg_QuiteException;
        fillchar(histlimg^,siv_histl_xsize*siv_hist_ysize*3,0);
        k1:=(siv_histl_xsize-siv_histr_xsize) shr 1;
        k2:=k1+siv_histr_xsize;

        j:=2;
        f:=histlc+ (k1+2)*img16.nc;
        for i:=2 to siv_histr_xsize-3 do
        begin
            if f[red]  >j then j:=f[red];
            if f[green]>j then j:=f[green];
            if f[blue] >j then j:=f[blue];
            f+=img16.nc;
        end;

        if log then r:=(siv_hist_ysize-3)/ln(j)
        else r:=(siv_hist_ysize-3)/j;
        for i:=0 to 2 do
        begin
            case i of
                0  : j:=red;
                1  : j:=green;
                else j:=blue;
            end;
            f:=histlc+j;
            for k:=0 to siv_histl_xsize-1 do
            begin
                if log then
                begin
                    l:=f^;
                    if l<1 then l:=siv_hist_ysize
                    else l:=siv_hist_ysize-1-round(ln(l)*r);
                end
                else l:=siv_hist_ysize-round(f^*r);
                if l<0 then l:=0;
                e:=histlimg + (k+l*siv_histl_xsize)*3+i;
                if (k<k1) or (k>=k2) then
                    for l:=l to siv_hist_ysize-1 do
                    begin
                        e^:=85; e+=siv_histl_xsize*3;
                    end
                else for l:=l to siv_hist_ysize-1 do
                begin
                    e^:=170; e+=siv_histl_xsize*3;
                end;
                f+=img16.nc;
            end;
        end;

        if not inv then j:=siv_hist_ysize-1
        else j:=0;
        histtrans[siv_histr_xsize]:=j;
        e:=histlimg + j*siv_histl_xsize*3;
        for i:=0 to k1 do
        begin
            e[0]:=255; e[1]:=255; e[2]:=255;
            e+=3;
        end;
        e:=histlimg + (siv_hist_ysize-1-j)*siv_histl_xsize*3 + k2*3;
        for i:=k2 to siv_histl_xsize-1 do
        begin
            e[0]:=255; e[1]:=255; e[2]:=255;
            e+=3;
        end;
        if not inv then
            for i:=0 to siv_histr_xsize do
            begin
                k:=siv_hist_ysize-1-histtrans[i];
                e:=histlimg + k*siv_histl_xsize*3 + (i+k1)*3;
                for l:=k to j do
                begin
                    e[0]:=255; e[1]:=255; e[2]:=255;
                    e+=siv_histl_xsize*3;
                end;
                j:=k;
            end
            else
                for i:=0 to siv_histr_xsize do
                begin
                    k:=siv_hist_ysize-1-histtrans[i];
                    e:=histlimg + j*siv_histl_xsize*3 + (i+k1)*3;
                    for l:=j to k do
                    begin
                        e[0]:=255; e[1]:=255; e[2]:=255;
                        e+=siv_histl_xsize*3;
                    end;
                    j:=k;
                end;
    except
        SIV_OutputCatchedException;
        fillchar(histlimg^,siv_histl_xsize*siv_hist_ysize*3,0);
    end;
//    siv_debug_end(-1,'CSIV_Image.drawhistl: ','performance: '+real2str((stimer-time)*1000,-2)+'ms ');
end;


{ ******** drawhist8 ****************************************************** }
procedure CSIV_Image.drawhistr;
var i,j,k,l,s       : longint;
    r               : double;
    d               : ^longint;
    e               : ^byte;
//    time            : double;
begin
//    siv_debug_begin(-1,'CSIV_Image.drawhistr','');
//    time:=stimer;
    try
        if histrc=nil then SImg_QuiteException;
        fillchar(histrimg^,siv_histr_xsize*siv_hist_ysize*3,0);
        s:=2;
        d:=histrc + 2*img16.nc;
        for i:=2 to siv_histr_xsize-3 do
        begin
            if d[red]  >s then s:=d[red];
            if d[green]>s then s:=d[green];
            if d[blue] >s then s:=d[blue];
            d+=img16.nc;
        end;

        if log then r:=(siv_hist_ysize-3)/ln(s)
        else r:=(siv_hist_ysize-3)/s;
        for i:=0 to 2 do
        begin
            case i of
                0  : j:=red;
                1  : j:=green;
                else j:=blue;
            end;
            d:=histrc+j;
            for k:=0 to siv_histr_xsize-1 do
            begin
                if log then
                begin
                    s:=d^;
                    if s<0.01 then l:=siv_hist_ysize
                    else l:=siv_hist_ysize-1-round(ln(s)*r);
                end
                else l:=siv_hist_ysize-round(d^*r);
                if l<0 then l:=0;
                e:=histrimg + (k+l*siv_histr_xsize)*3+i;
                for l:=l to siv_hist_ysize-1 do
                begin
                    e^:=170; e+=siv_histr_xsize*3;
                end;
                d+=img16.nc;
            end;
        end;
    except
        SIV_OutputCatchedException;
        fillchar(histrimg^,siv_histr_xsize*siv_hist_ysize*3,0);
    end;
//    siv_debug_end(-1,'CSIV_Image.drawhistr: ','performance: '+real2str((stimer-time)*1000,-2)+'ms ');
end;


{ ******** history_markers ************************************************ }
procedure CSIV_Image.history_markers_int;
var i,j,k,l       : longint;
    so,sr,r65535  : double;
    d             : ^word;
    e             : ^byte;
begin
    if (marker_x>=0) and (marker_y>=0) and (marker_x<img16.nx) and (marker_y<img16.ny) then
    begin
        r65535:=1/65535;
        i:=round(ll); if i<0 then i:=0; if i>65535 then i:=65535;
        j:=round(ul); if j<0 then j:=0; if j>65535 then j:=65535;
        if abs(j-i)<2 then siv_error('Limits to close');

        so:=i/(1-delta+delta*i*r65535);
        sr:=(siv_histr_xsize-1)/(j/(1-delta+delta*j*r65535)-so);

        d:=img16.ptw + marker_y*img16.nx*img16.nc + marker_x*img16.nc;

        for i:=0 to 2 do
        begin
            case i of
                0  : j:=d[red];
                1  : j:=d[green];
                else j:=d[blue];
            end;

            k:=((siv_histl_xsize-siv_histr_xsize) shr 1) + round((j/(1-delta+delta*i*r65535)-so)*sr);
            if k<0 then k:=0;
            if k>=siv_histl_xsize then k:=siv_histl_xsize-1;

            e:=histlimg+k*3+i;
            for l:=0 to siv_hist_ysize-1 do
            begin
                if e^<255 then e^:=254-e^;
                e+=siv_histl_xsize*3;
            end;

            k:=(xlat[j]*siv_histr_xsize) shr 8;
            e:=histrimg+k*3+i;
            for l:=0 to siv_hist_ysize-1 do
            begin
                e^:=255-e^;
                e+=siv_histr_xsize*3;
            end;
        end;
    end;
end;


procedure CSIV_Image.history_markers(x,y:longint);
begin
    if (marker_x<>x) or (marker_y<>y) then
    begin
        history_markers_int;
        marker_x:=x;
        marker_y:=y;
        history_markers_int;
    end;
end;

